# -*- coding: utf-8 -*-
"""Dependency tests for bran."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

def test_pyasn1_version():
  from pyasn1 import __version__
  major, minor, patch = [int(x) for x in __version__.split('.')]
  assert major == 0
  assert minor >= 4



# -*- coding: utf-8 -*-
"""Test suite for bran."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

@pytest.fixture(scope = 'module')
def nested_data():
  theset = set()
  theset.add(1)
  theset.add(2)
  theset.add(1)

  from collections import OrderedDict
  x = OrderedDict()
  x['baz'] = 42
  x['foo'] = 'bar'

  return {
    'set': theset,
    'none': None,
    'numbers': {
      'int': 42,
      'float': 3.1415,
      'complex': complex(1, -2),
    },
    'list': ['y', 'n'],
    'bools': (True, False),
    'strings': ("hällo",u"hällo",b"hallo"),
    'dict': x,
  }



# -*- coding: utf-8 -*-
"""
This module provides a simple object hashing abstraction.

We use bran.DERTranscoder for serialization, and a hash function
from hashlib.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import hashlib


def hasher(obj = None, hashfunc = hashlib.sha512, *args, **kwargs):
  """
  Create a hashlib wrapper that hashes objects.

  The returned object implements `update`, `digest` and `hexdigest` like
  the hashblib hash functions. The difference is that instead of only
  accepting buffer API objects in `update`, any object that can be serialized
  using DERTranscoder is supported.

  :param mixed obj: An optional object to update the hash function with.
  :param callable hashfunc: One of hashlib's constructor functions; defaults to
    hashlib.sha512
  :return: A hashlib-like hasher.
  """
  class BranHasher(object):
    def __init__(self, obj, *args, **kwargs):
      # Initialize chosen hash function
      self.__hashfunc = hashfunc(*args, **kwargs)

      # Initialize transcoder
      from . import DERTranscoder
      self.__transcoder = DERTranscoder()

      # Start hashing if we've been given an object in the ctor
      if obj is not None:
        self.update(obj)

    def update(self, *args, **kwargs):
      # Replace args by encoded versions of its objects.
      encoded = [self.__transcoder.encode(x) for x in args]
      return self.__hashfunc.update(*encoded, **kwargs)

    def digest(self, *args, **kwargs):
      return self.__hashfunc.digest(*args, **kwargs)

    def hexdigest(self, *args, **kwargs):
      return self.__hashfunc.hexdigest(*args, **kwargs)

  return BranHasher(obj, *args, **kwargs)



# -*- coding: utf-8 -*-
"""Test suite for bran.ASN1Transcoder."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

@pytest.fixture
def transcoder():
  from bran import ASN1Transcoder
  return ASN1Transcoder()

class Foo(object):
  def __eq__(self, other):
    return self.__dict__ == other.__dict__


def set_testing(transcoder, tag, values):
  from bran.util import stringify
  for value in values:
    # Encoding must result in the expected tag set
    encoded = transcoder.encode(value)
    assert stringify(encoded.getTagSet()) == tag

    # Decoding must restore the value
    decoded = transcoder.decode(encoded)
    assert decoded == value

    # The value *type* must be derived from the decoded type (i.e. we can
    # lose some information here)
    assert isinstance(value, type(decoded))


def test_none(transcoder):
  set_testing(transcoder, '[0:0:5]', (None, ))


def test_boolean(transcoder):
  set_testing(transcoder, '[0:0:1]', (True, False))


def test_int(transcoder):
  set_testing(transcoder, '[0:0:2]', (-1, 0, 42, 2147483648))


def test_float(transcoder):
  set_testing(transcoder, '[0:0:9]', (-1.0, 0.0, 42.0, 2147483648.0))


def test_complex(transcoder):
  set_testing(transcoder, '[0:32:16]+[128:32:1]', (complex(1, 2), complex(complex(1, 2), -3.14)))


def test_binstr(transcoder):
  set_testing(transcoder, '[0:0:4]', ("test".encode('ascii'), ))


def test_bytes(transcoder):
  set_testing(transcoder, '[0:0:4]', (b'testvalue', ))


def test_text(transcoder):
  set_testing(transcoder, '[0:0:12]', (u"hello!", ))


def test_dict(transcoder):
  from collections import OrderedDict
  # String keys
  set_testing(transcoder, '[0:32:16]+[128:32:4]', ({'a':2}, OrderedDict({'a':2})))
  # Numeric keys
  set_testing(transcoder, '[0:32:16]+[128:32:4]', ({1:2}, OrderedDict({1:2})))


def test_set(transcoder):
  val = set()
  val.add(1)
  val.add(2)
  set_testing(transcoder, '[0:32:17]+[128:32:5]', (val, ))


def test_tuple(transcoder):
  set_testing(transcoder, '[0:32:16]+[128:32:2]', ((1, 'a'), ))


def test_list(transcoder):
  set_testing(transcoder, '[0:32:16]+[128:32:3]', ([1, 'a'], ))


def test_dict_sorting():
  # OrderedDict keeps keys in insertion order, so the following two are
  # distinct.
  from collections import OrderedDict
  x = OrderedDict()
  x['foo'] = 'bar'
  x['baz'] = 42
  y = OrderedDict()
  y['baz'] = 42
  y['foo'] = 'bar'
  assert x != y

  # Default encoding will sort by keys, so would make the dicts equal
  from bran import ASN1Transcoder

  tc1 = ASN1Transcoder()
  x1 = tc1.encode(x)
  y1 = tc1.encode(y)
  assert x1 == y1

  # With sorting turned off, they should not be.
  tc2 = ASN1Transcoder(sort = False)
  x2 = tc2.encode(x)
  y2 = tc2.encode(y)
  assert x2 != y2

  # Try a sorting function. If we use an inverting function, y put
  # through the transcoder should equal x
  tc3 = ASN1Transcoder(sort = reversed)
  y3 = tc3.decode(tc3.encode(y))
  assert x == y3


def test_nested(transcoder, nested_data):
  decoded = transcoder.decode(transcoder.encode(nested_data))
  assert decoded == nested_data


def test_registry_basics():
  from bran import ASN1Transcoder

  test = { 'foo': Foo() }

  # Without a converter registry, we can't encode this
  with pytest.raises(TypeError):
    tc = ASN1Transcoder()
    encoded = tc.encode(test)

  from pyasn1.type import univ

  registry = {
    Foo: lambda x: univ.ObjectIdentifier([42]),
    '[0:0:6]': lambda x: Foo()
  }

  # With registry, we don't expect errors
  tc = ASN1Transcoder(registry = registry)
  encoded = tc.encode(test)

  # Decoding should fail without registry
  with pytest.raises(TypeError):
    tc1 = ASN1Transcoder()
    decoded = tc1.decode(encoded)

  # But with registry it must succeed
  decoded = tc.decode(encoded)
  assert decoded == test


def test_registry_no_callable():
  from bran import ASN1Transcoder

  registry = {
    Foo: 'not a callable',
    '[0:0:6]': 'not a callable',
  }
  tc = ASN1Transcoder(registry = registry)

  test = { 'foo': Foo() }
  with pytest.raises(ValueError):
    encoded = tc.encode(test)

  with pytest.raises(ValueError):
    from pyasn1.type import univ
    decoded = tc.decode(univ.ObjectIdentifier([42]))




# -*- coding: utf-8 -*-
"""
Utility code for bran.

Contains a stringify function for tags/tagsets that does not include object
IDs, and thus becomes comparable.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()


def stringify(tag_or_tagset):
  """
  Return a string representation of a Tag or TagSet.

  Tag or TagSet are from pyasn1.type.tag.

  :param mixed tag_or_tagset: Either a Tag or a TagSet instance.
  :return: String representation.
  :raises: TypeError if the value is not a Tag or TagSet.
  """
  from pyasn1.type.tag import Tag, TagSet
  if isinstance(tag_or_tagset, Tag):
    return '[{0}:{1}:{2}]'.format(
        tag_or_tagset.tagClass,
        tag_or_tagset.tagFormat,
        tag_or_tagset.tagId)
  elif isinstance(tag_or_tagset, TagSet):
    return '+'.join(stringify(tag) for tag in tag_or_tagset)
  raise TypeError('{0} is not an instance of Tag or TagSet'.format(
      str(tag_or_tagset)))



# -*- coding: utf-8 -*-
"""Test suite for bran.util."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

def test_stringify():
  from pyasn1.type.tag import Tag, TagSet
  from bran.util import stringify

  # First a single tag
  tag = Tag(0, 1, 2)
  assert '[0:1:2]' == stringify(tag)

  # Now a tag set with a single tag
  tagset = TagSet() + tag
  assert '[0:1:2]' == stringify(tagset)

  # With multiple tags
  tagset = TagSet() + tag + tag
  assert '[0:1:2]+[0:1:2]' == stringify(tagset)

  # With invalid value
  with pytest.raises(TypeError):
    stringify(None)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Dist utils for Bran.

This should work within tox, and install all required dependencies for testing.
"""

if __name__ == '__main__':
  # Get setup function
  try:
    from setuptools import setup, find_packages
  except ImportError:
    from distutils.core import setup, find_packages

  dev_require = [
    'tox>=2.8',
    'bump2version>=0.5',
    'pytest>=3.2',
    'pytest-cov>=2.5',
    'flake8>=3.4',
    'pep8-naming>=0.4',
    'flake8-quotes>=0.11',
    'flake8_docstrings>=1.1',
    'sphinx>=1.6',
  ]

  # Run setup
  setup(
      name = 'bran',
      version = '0.4.0',
      description = 'DER-Encoded ASN.1 Serialization and Deserialization',
      long_description = open('README.rst').read(),
      # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords = 'ASN.1 DER serialization',
      author = 'Jens Finkhaeuser',
      author_email = 'jens@finkhaeuser.de',
      url = 'https://github.com/jfinkhaeuser/bran',
      license = 'MITNFA',
      packages = find_packages(exclude = ['ez_setup', 'examples', 'tests']),
      include_package_data = True,
      install_requires = [
        'six~=1.11',
        'pyasn1~=0.4',
      ],
      extras_require = {
        'dev': dev_require,
      },
      scripts = [
      ],
      zip_safe = True,
      test_suite = 'tests',
      setup_requires = ['pytest-runner'],
      tests_require = dev_require,
  )



# -*- coding: utf-8 -*-
"""Test suite for bran.DERTranscoder."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

@pytest.fixture
def transcoder():
  from bran import DERTranscoder
  return DERTranscoder()


def test_nested(transcoder, nested_data):
  decoded = transcoder.decode(transcoder.encode(nested_data))
  assert decoded == nested_data




# -*- coding: utf-8 -*-
"""Test suite for bran.hash."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ()

import pytest

def test_hash():
  # OrderedDict keeps keys in insertion order, so the following two are
  # distinct.
  from collections import OrderedDict
  x = OrderedDict()
  x['foo'] = 'bar'
  x['baz'] = 42
  y = OrderedDict()
  y['baz'] = 42
  y['foo'] = 'bar'
  assert x != y

  # Hashing will sort keys, so the two hash digests must be the same.
  from bran.hash import hasher

  hx = hasher(x).hexdigest()
  hy = hasher(y).hexdigest()

  assert hx == hy

  hx = hasher(x).digest()
  hy = hasher(y).digest()

  assert hx == hy



# -*- coding: utf-8 -*-
"""
Bran provides DER-Encoded ASN.1 Serialization and Deserialization.

Other serialization methods abound, but DER-encoded ASN.1 is used in
cryptography for one particular reason: given the same object, its encoding
is the same, which means hashes over its encoded form are.

For much the same reason, DER-encoded ASN.1 is useful in other contexts.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2018 Jens Finkhaeuser'
__license__ = 'MIT +no-false-attribs'
__all__ = ('hash')
__version__ = '0.4.0'


class DERTranscoder(object):
  """
  DER-encode Python builtin (and extended) types.

  The class first uses ASN1Transcoder to encode valuesin ASN.1 classes, then
  DER-encodes the result.
  """

  def __init__(self, inner = None):
    """
    Initialize DERTranscoder.

    The sole parameter is a reference to an inner decoder. If not specified,
    this defaults to an ASN1Transcoder instance with default options.

    :param object inner: [optional] Inner transcoder
    """
    self.inner = inner or ASN1Transcoder()

  def encode(self, value):
    """
    DER-encode the given value.

    Nested values are encoded recursively.

    :param mixed value: The value to encode.
    :return: An DER-encoded ASN.1 value, i.e. a byte sequence.
    """
    from pyasn1.codec.der import encoder
    return encoder.encode(self.inner.encode(value))

  def decode(self, value):
    """
    DER-deocde the given byte sequence.

    :param bytes value: The value to decode.
    :return: A Python value, the result of passing the parameter through DER
        decoding and ASN.1 decoding.
    """
    from pyasn1.codec.der import decoder
    decoded = decoder.decode(value)
    return self.inner.decode(decoded[0])


class ASN1Transcoder(object):
  """
  Transcode Python builtin (and extended) types to and from ASN.1 classes.

  Note that encoding is lossy with regards to specialized types.

  For many Python builtin types, there are corresponding universal ASN.1 data
  types available.

  For tuple() and list(), we use explicit tagging (which, when DER-encoded,
  takes up a few extra Bytes of space) to disambiguate one kind of sequence
  from another. Any collections.Sequence that is not specifically a list is
  encoded as a tuple(), any additional type information.

  Similarly, any collections.Mapping is encoded in the same way, and so is
  any collections.Set. If you use special subtypes of these Abstract Base
  Classes, that type information is lost.

  Still, it means that any nested structure will result in an equality-
  comparable nested structure, which is all that often times is required.
  """

  def __init__(self, **kwargs):
    """
    Initialize the transcoder.

    Keyword arguments are stored as options. Some such recognized options are:

    :param bool sort: Either a key sorting function, or False. If False, no
        sorting is performed. The default is to perform key sorting using the
        sorted() function.
    :param Mapping registry: A collections.Mapping, with keys specifying either
        Python types (not names/strings, but the type objects), or ASN.1 tags
        (stringified, e.g. "[0:0:6]"). The value must be a callable that
        converts from objects of the Python type to an ASN.1 object or from an
        ASN.1 object matching the stringified tag to a Python object.
    """
    self.options = kwargs

    from pyasn1.type import univ, tag

    # Tags for complex
    self.COMPLEX = univ.Sequence(
      tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0x01)
      )
    )
    # Tags for collections.Sequence
    self.TUPLE = univ.Sequence(
      tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0x02)
      )
    )
    self.LIST = univ.Sequence(
      tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0x03)
      )
    )
    # Tags for collections.Mapping
    self.MAPPING = univ.Sequence(
      tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0x04)
      )
    )
    # Tags for collections.Set
    self.SET = univ.Set(
      tagSet = univ.Set.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0x05)
      )
    )

  def encode(self, value):
    """
    Encode the given Python value.

    Nested values (for Sequence, Set and Mapping) are recursively encoded.

    :param mixed value: The value to encode.
    :return: An ASN.1 class encapsulating the value.
    """
    import six, collections
    from pyasn1.type import univ, char

    if value is None:
      return univ.Null()

    elif isinstance(value, bool):
      return univ.Boolean(value)

    elif isinstance(value, six.integer_types):
      return univ.Integer(value)

    elif isinstance(value, float):
      return univ.Real(value)

    elif isinstance(value, complex):
      return self.__encode_complex(value)

    elif isinstance(value, (bytes, bytearray, six.binary_type)):
      return univ.OctetString(bytearray(value))

    elif isinstance(value, six.text_type):
      return char.UTF8String(value.encode('utf8'))

    elif isinstance(value, collections.Mapping):
      return self.__encode_mapping(value)

    elif isinstance(value, collections.Set):
      return self.__encode_set(value)

    elif isinstance(value, collections.Sequence):
      return self.__encode_sequence(value)

    else:
      return self.__encode_from_registry(value)

  def __encode_from_registry(self, value):
    registry = self.options.get('registry', {})

    # If we find the value class in the registry, we can use that to create
    # a custom ASN.1 type.
    klass = type(value)
    encoder = registry.get(klass, None)
    if encoder is None:
      raise TypeError('Cannot encode value of type "%s"!' % (klass,))

    if not callable(encoder):
      raise ValueError('Bad registry entry "%s" for class "%s"; expect a '
          'callable!' % (encoder, klass))

    # First call is for encoding
    return encoder(value)

  def __encode_complex(self, value):
    # Encode complex() values
    val = self.COMPLEX.clone()

    val[0] = self.encode(value.real)
    val[1] = self.encode(value.imag)

    return val

  def __encode_sequence(self, value):
    # Everything except for lists get coerced to a tuple
    base = self.TUPLE
    if isinstance(value, list):
      base = self.LIST
    val = base.clone()

    # Sequences can't be re-ordered
    for idx, item in enumerate(value):
      val[idx] = self.encode(item)

    return val

  def __encode_mapping(self, value):
    # coerce to dict
    val = self.MAPPING.clone()

    # Force ordering of keys by default. Turn keys into a list for Python 3.4
    # compatibility.
    keys = list(value.keys())
    sorter = self.options.get('sort', sorted)
    if sorter is not False:
      keys = sorter(keys)

    for idx, key in enumerate(keys):
      val[idx] = self.encode((key, value[key]))

    return val

  def __encode_set(self, value):
    # Everything is a set
    val = self.SET.clone()

    # Force ordering of items by default
    items = value
    sorter = self.options.get('sort', sorted)
    if sorter is not False:
      items = sorter(items)

    for idx, item in enumerate(items):
      val[idx] = self.encode(item)

    return val

  def decode(self, value):
    """
    Decode the given ASN.1 class into a Python value.

    Nested values (for Sequence, Set and Mapping) are recursively decoded.

    :param mixed value: The value to decode.
    :return: An Python value.
    """
    # import six, collections
    from pyasn1.type import univ, char

    if isinstance(value, univ.Null):
      return None

    elif isinstance(value, univ.Boolean):
      return bool(value)

    elif isinstance(value, univ.Integer):
      return int(value)

    elif isinstance(value, univ.Real):
      return float(value)

    elif isinstance(value, char.UTF8String):
      from six import text_type
      return text_type(value)

    elif isinstance(value, univ.OctetString):
      from six import binary_type
      return binary_type(value)

    elif isinstance(value, (univ.Sequence, univ.SequenceOf)):
      # Look for sub type tags
      if value.isSameTypeWith(self.COMPLEX):
        return complex(
            self.decode(value[0]),
            self.decode(value[1])
        )

      if value.isSameTypeWith(self.TUPLE):
        return tuple(self.__sequence_iter(value))

      if value.isSameTypeWith(self.LIST):
        return list(self.__sequence_iter(value))

      if value.isSameTypeWith(self.MAPPING):
        ret = {}
        for key, value in self.__sequence_iter(value):
          ret[key] = value
        return ret

    elif isinstance(value, (univ.Set, univ.SetOf)):
      ret = set()
      for item in self.__sequence_iter(value):
        ret.add(item)
      return ret

    else:
      return self.__decode_from_registry(value)

  def __decode_from_registry(self, value):
    registry = self.options.get('registry', {})

    # If we find the value tagset in the registry, we can use that to create
    # a custom ASN.1 type.
    from .util import stringify
    tagset = stringify(value.tagSet)
    decoder = registry.get(tagset, None)
    if decoder is None:
      raise TypeError('Cannot decode value tag set "%s"!' % (tagset,))

    if not callable(decoder):
      raise ValueError('Bad registry entry "%s" for tag set "%s"; expect a '
          'callable!' % (decoder, tagset))

    # First call is for encoding
    return decoder(value)

  def __sequence_iter(self, seq):
    # Iterate through a univ.Sequence, decode and yield each item
    idx = 0
    while True:
      try:
        item = self.decode(seq[idx])
        yield item
        idx += 1
      except IndexError:
        break
    raise StopIteration
